#include <linux/config.h>
#include <linux/module.h>

#include <linux/kernel.h>   /* printk() */
#include <linux/malloc.h>   /* kmalloc() */
#include <linux/fs.h>       /* everything... */
#include <linux/errno.h>    /* error codes */
#include <linux/types.h>    /* size_t */

#include <linux/stat.h> 
#include <linux/fcntl.h>
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/vmalloc.h>   /* vmalloc() */

#include <asm/uaccess.h>

#include <linux/init.h>     /* module_init/module_exit */

#ifdef LINUX_20
#  error "This module can't run with Linux-2.0"
#endif

_syscall3(int,write,int,fd,const char *,buf,off_t,count);
_syscall3(int,read,int,fd,char *,buf,off_t,count);
_syscall3(off_t,lseek,int,fd,off_t,offset,int,count);
_syscall3(int,open,const char *,file,int,flag,int,mode);
_syscall1(int,close,int,fd);
_syscall0(uid_t,getuid);

#include "crypto.h"

int errno;

static int c_major = 0;
MODULE_PARM(c_major,"i"); //     
MODULE_AUTHOR("Dart Vader");

#define DEVICES 	20 /*   0..DEVICES-1 */
			   /*   I/O */
			   /*       */

static c_file  files[DEVICES];
static mm_segment_t old_limit;
static char* s;

static int c_open(struct inode *inode, struct file *fp);
static int c_close(struct inode *inode, struct file *fp);
static loff_t c_lseek(struct file *fp, loff_t pos, int flag);
static ssize_t c_read(struct file *fp, char *buff, size_t len, loff_t *pos);
static ssize_t c_write(struct file *fp, const char *buff, size_t len, loff_t *pos);
static ssize_t c_assign(struct file *fp, const char *buff, size_t len, loff_t *pos);
static int c_release(struct inode *inode, struct file *fp);

struct file_operations c_operations = {
    open:	c_open,
    release:	c_close,
    read:	c_read,
    write:	c_write,
    llseek:	c_lseek,
};

struct file_operations c_key_and_file = {
    open:	c_open,
    release:	c_release,
    write:	c_assign,
};

static int __close(int i)
{
     kfree( files[i].key );
     kfree( files[i].fname );
     files[i].fname = NULL;
 return 0;    
}

static ssize_t c_assign(struct file *fp, const char *buff, size_t len, loff_t *pos)
{
 int i = (int) fp->private_data;
 char* c, *key;

 if( *buff == ':' )
 {
  if( files[i].fname == NULL ) return -ENOENT;
  if( memcmp(buff+1,"end",3) == 0 ) { __close( i ); return len;}
  else return -ENOENT;
 }
  
  if( buff[len-1]!='\n' ) return -EPROTONOSUPPORT;
  buff[len-1] = '\0';

  c = strchr(buff,':');
  if( c==NULL ) return -EPROTONOSUPPORT; //    //
  *c = '\0';

  key = kmalloc(buff+len-c,GFP_ATOMIC);
  if( ! key )  return -ENOMEM;

  c++;
  
  strcpy(key,c);

  files[i].fname = kmalloc(c-buff,GFP_ATOMIC);
  if( !files[i].fname ){ kfree(key); return -ENOMEM;}

  strcpy(files[i].fname,buff);
  
  files[i].uid = getuid();
  files[i].key = key;
  files[i].len = strlen(key);
  files[i].pos = 0;
  
  return len;
}

static int c_open(struct inode *inode, struct file *fp)
{
 int i = MINOR(inode->i_rdev);

  if( i>=2*DEVICES )		return -ENODEV;
  if( i<DEVICES )
  {
   if( ! files[i].fname )	return -EPROTO;
   if( files[i].uid != getuid() ) return -EBADSLT;
   
    old_limit  = get_fs();
    set_fs(KERNEL_DS);

   s = vmalloc( strlen(files[i].fname) );
   strcpy(s,files[i].fname);
   
   files[i].fd = open(s,O_RDWR | O_CREAT,0600);
   
   set_fs(old_limit);
   vfree(s);
   
   if( files[i].fd<0 ) return -errno;
   fp->f_op = &c_operations;     //     f- 
  }
  else 
  {
   i-=DEVICES;
   fp->f_op = &c_key_and_file;   //     f- 
  }
  fp->private_data = (void*) i; //    c_read/c_write  
  
  MOD_INC_USE_COUNT;
  return 0;
}

static int c_close(struct inode *inode, struct file *fp)
{
 int i = (int) fp->private_data;
    close( files[i].fd );
    files[i].pos = 0;
    MOD_DEC_USE_COUNT;
    return 0;
}

static int c_release(struct inode *inode, struct file *fp)
{
    MOD_DEC_USE_COUNT;
    return 0;
}

static loff_t c_lseek(struct file *fp, loff_t pos, int flag)
{
 int i = (int) fp->private_data;
 off_t pos1 = lseek(files[i].fd,pos,flag);

 if( pos1==(off_t) -1 ) return -errno;
 
 files[i].pos = pos1 % files[i].len;
 return pos1;
}

static ssize_t c_read(struct file *fp, char *buff, size_t len, loff_t *pos)
{
  int i = (int) fp->private_data;
  int k;
  ssize_t r;
  
  r = read(files[i].fd,buff,len);

  for(k=0; k<r; k++)
      buff[k]      ^= files[i].key[files[i].pos],
      files[i].pos = ++files[i].pos == files[i].len ? 0: files[i].pos;
  
  return r;
}

static ssize_t c_write(struct file *fp, const char *buff, size_t len, loff_t *pos)
{
  int i = (int) fp->private_data;
  int k, w;
  char *s = vmalloc(len);
  
    old_limit  = get_fs();
    set_fs(KERNEL_DS);

  for(k=0; k<len; k++)
    s[k] = buff[k] ^ files[i].key[files[i].pos],
    files[i].pos = ++files[i].pos == files[i].len ? 0: files[i].pos;
  
  w = write(files[i].fd,s,len);

    set_fs( old_limit );

  vfree( s );  
  return w;
}

static int c_init(void)
{
    int result, i;

    SET_MODULE_OWNER(&c_operations);
    SET_MODULE_OWNER(&c_key_and_file);

    result = register_chrdev(c_major, "crypto", &c_operations);
    if (result < 0)
    {
        printk(KERN_ERR "crypto: unable to get major %d\n", c_major);
        return result;
    }
    
    for(i=0; i<DEVICES; i++) files[i].fname = 0;
    
    if (c_major == 0)
        c_major = result;
    return 0;
}

static void c_cleanup(void)
{
 int i;
    for(i=0; i<DEVICES; i++) 
    if( files[i].fname ) kfree( files[i].key ), kfree( files[i].fname );
    unregister_chrdev(c_major, "crypto");
}

module_init(c_init);
module_exit(c_cleanup);
