/* edkey.c
 *
 *  /      edcrypt.
 *     ioct .
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <getopt.h>
#include <stdarg.h>

#include "edcrypt.h"


static void
usage(void)
{
    fprintf(stderr, "edkey - change secret key (password) on\n\
            /dev/edcrypt and /dev/decrypt special files\nusage:\n\
            edkey [ -f special_file ] [ -d ] \n\
            -d                 # delete current encryption key info from driver\n\
            -f special_file    # use another device\n\
			       # without parameters - enter new encryption key.\n");
    exit(1);
}

void
xstrncpy(char *dest, const char *src, size_t n)
{
    strncpy(dest, src, n-1);
    dest[n-1] = 0;
}

int
main(int argc, char **argv)
{
    char device[255] = "/dev/encrypt";
    char *pass;
    int c, fd, delete = 0;
    struct encrypt_key_info new_key;

    while ((c = getopt(argc,argv,"df:h")) != EOF)
    {
        switch (c)
        {
        case 'd':
            delete = 1;
            break;
        case 'f':
            strncpy(device, optarg, sizeof(device));
            break;
        default:
            usage();
        }
    }

    if ((fd = open (device, O_RDONLY)) < 0)
    {
        int errsv = errno;
        fprintf(stderr, "edcrypt: can't open device %s: %s\n",
                device, strerror (errsv));
        return 2;
    }

    memset (&new_key, 0, sizeof(new_key));

    if (!delete)
    {
        pass = getpass ("Password: ");
        xstrncpy (new_key.key, pass,  EDCRYPT_KEY_SIZE);
        new_key.size = strlen(new_key.key);
    }

//    printf("New key[%s], %d\n", new_key.key, new_key.size);

    if (ioctl (fd, EDCRYPT_SET_KEY, &new_key) < 0)
    {
        perror ("ioctl: EDCRYPT_SET_KEY");
        close(fd);
        return 1;
    }
    close (fd);
    fprintf(stdout, "edkey: password key store success :)\n");

    return 0;
}
